/* Adaptive Huffman encoding program */
/* Hal Burch */
/* Sept 16, 1996 */

/* This is based on the Adaptive Huffman algorithm described in Sayood's Data
 * Compression book.  The ranks are not actually stored, but implicitly defined
 * by the location of a node within a doubly-linked list */


/*******************************************************************************
*NOTICE:                                                                       *
*This code is believed by the author to be bug free.  You are free to use and  *
*modify this code with the understanding that any use, either direct or        *
*derivative, must contain acknowledgement of its author and source.  The author*
*makes no warranty of any kind, expressed or implied, of merchantability or    *
*fitness for a particular purpose.  The author shall not be held liable for any*
*incidental or consequential damages in connection with or arising out of the  *
*furnishing, performance, or use of this software.  This software is not       *
*authorized for use in life support devices or systems.                        *
********************************************************************************/
/* Modified by David A. Scott so could be better used for cryptograpic uses */
/* on 2000 01 10 */
/* #define ASSERT 1 /**/
#define MAX 255 /* Maximum symbol */
#define N 8 /* number of bits */

#include <stdio.h>
#include <assert.h>
#include <malloc.h>
#include <string.h>
#include "ad2hc.h"

extern Node *tree;
extern Node *lhead;
extern Node *loc[MAX+1];
static  Node **freelist=NULL;

Node **get_ppnode(void)
 {
  Node **tppnode;
  if (!freelist)
    return (Node **)malloc(sizeof(Node *));
  else
   {
    tppnode = freelist;
    freelist = (Node **)*tppnode;
    return tppnode;
   }
 }

static void free_ppnode(Node **ppnode)
 {
  *ppnode = (Node *)freelist;
  freelist = ppnode;
 }

void checkranks(Node *node)
 {
  Node *lnode;
  for (lnode = node; lnode; lnode = lnode->next) 
   {
 /* printf(" check %d %d \n",(int)(*lnode->head)->weight,(int)lnode->weight);*/
    assert(!lnode->next || lnode->next->weight >= lnode->weight);
    assert(lnode->head && *(lnode->head) && (*lnode->head)->weight == lnode->weight);
    if (lnode->next)
      if (lnode->weight == lnode->next->weight)
        assert(lnode->next->head == lnode->head);
      else
        assert(lnode->next->head != lnode->head);
   }
 }
static int KOUNT;
void dave(Node *node ){
  if ( !node->parent) return;
  KOUNT=0;
  dave ( node->parent);
  KOUNT++;
  if ( node->parent->left == node) printf("l");
  else if ( node->parent->right == node) printf("r");
  else { printf(" TREE ERROR \n");exit(0);}
}
void scott(Node *node){
 Node *das,*dv; /* dv for Daneil.Vogelheim */
 int i,j;
 for ( i= 0,dv = lhead; dv; dv = dv->parent,i++ ){
  if ( dv->parent->left != dv ) {
    dv->parent->right = dv->parent->left;
    dv->parent->left = dv; 
  }
    if( i > 300 ) break; 
}
if ( i > 300 ){
 printf (" **fatal DV error ** i = %d \n",i);
 exit (0);
 }
 
 if( node->right && node->left){
 for(i=0,das=node->right;das=das->right;)i++;
 for(j=0,das=node->left;das=das->left;)j++;
  if (i > j  || ( i=j && node->right < node->left)){
 das=node->right;
 node->right=node->left;
node->left=das;} 
}}
/**/
void dumptree(Node *node)
 { /* Debugging routine...dump the tree in prefix notation */
  if (!node) return;
  if (node->symbol == INTERNAL_NODE);
   /* printf (".");*/
  else {
   printf ("z%2.2xz", node->symbol);
   printf("W%8.8xW ",node->weight);
   dave(node);
   printf(" bits= %d\n",KOUNT); 
}
  if (node->left)
    assert(node->left->parent == node);
  if (node->right)
    assert(node->right->parent == node);
  dumptree(node->left);
  dumptree(node->right);
  if (node == tree)
    printf ("\n");
 }
/* */
void dumplist(Node *node)
 { /* Debugging routine...dump the link list */
  while (node)
   {
    if (!node) return;
    if (node->symbol == INTERNAL_NODE)
      printf (".");
    else printf ("s%xs", node->symbol);
    printf ("(%i)", node->weight);
    node = node->next;
   }
  printf ("\n");
 }

void usage(char **argv)
 {
  fprintf (stderr, "Usage: %s inputfile outputfile\n", argv[0]);
 }

void swap (Node *node1, Node *node2)
 { /* Swap the location of these two nodes in the tree */
  Node *par1, *par2;


  par1 = node1->parent;
  par2 = node2->parent;

  if (par1)
   {
    if (par1->left == node1)
      par1->left = node2;
    else
     {
#ifdef ASSERT
      assert(par1->right == node1);
#endif
      par1->right = node2;
     }
   }
  else tree = node2;

  if (par2)
   {
    if (par2->left == node2)
      par2->left = node1;
    else
     {
#ifdef ASSERT
      assert(par2->right == node2);
#endif
      par2->right = node1;
     }
   }
  else
    tree = node1;
  
  node1->parent = par2;
  node2->parent = par1;
  scott(par1);
  scott(par2);
 }

void swaplist(Node *node1, Node *node2)
 { /* Swap these two nodes in the linked list (update ranks) */
  Node *par1;

  par1 = node1->next;
  node1->next = node2->next;
  node2->next = par1;

  par1 = node1->prev;
  node1->prev = node2->prev;
  node2->prev = par1;

  if (node1->next == node1)
    node1->next = node2;
  if (node2->next == node2)
    node2->next = node1;

  if (node1->next)
    node1->next->prev = node1;
  if (node2->next)
    node2->next->prev = node2;
  if (node1->prev)
    node1->prev->next = node1;
  if (node2->prev)
    node2->prev->next = node2;
#ifdef ASSERT
  assert(node1->next != node1);
  assert(node2->next != node2);
#endif
 }

void increment(Node *node,int DAS)
 { /* Do the increments */
  Node *lnode;

  if (!node) return;


  if (node->next != NULL && node->next->weight == node->weight)
   {
    lnode = *node->head;
#ifdef ASSERT
    assert(lnode != node);
    assert(lnode->parent != node);
    assert(lnode->weight == node->weight);
#endif
    if (lnode != node->parent)
      swap(lnode, node);
    swaplist(lnode, node);
   }

  if (node->prev && node->prev->weight == node->weight)
    *node->head = node->prev;
  else
   {
    *node->head = NULL;
    free_ppnode(node->head);
   }

   node->weight= DAS + node->weight; 
    /* scott(node); */
   

    
  if (node->next && node->next->weight < node->weight){
  printf(" *** errir \n %x %x %x\n",(int)node->next->weight,(int)node->weight,DAS); exit(8);}
  if (node->next && node->next->weight == node->weight)
    node->head = node->next->head;
  else
   { 
    node->head = get_ppnode();
    *node->head = node;
   }
  if (node->parent)
   {
    increment(node->parent,DAS);
    if (node->prev == node->parent)
     {
      swaplist(node, node->parent);
      if (*node->head == node)
        *node->head = node->parent;
     }
   }
 }
static int XXX[MAX+1]={0,0};
static int YYY=MAX+1;
 void add_ref(int ch)
  {
   int i;
   Node *tnode, *tnode2;
   if (!YYY)
       increment(loc[ch],4*(MAX+1));
    else if ( !XXX[ch])
      {
/*
     if (YYY <= 201)printf("here f \n");
     printf(" topV baby %x weight is %x \n",ch,(int)loc[ch]->weight);
     printf(" topV parent %x weight is %x \n",ch,(int)loc[ch]->parent->weight);
     printf(" topV lhead %x weight is %x \n",ch,(int)lhead->weight);
*/
       while (loc[ch]->weight < 2*(MAX+1)) increment(loc[ch],1);
/*
     printf(" topW baby %x weight is %x \n",ch,(int)loc[ch]->weight);
     printf(" topW parent %x weight is %x \n",ch,(int)loc[ch]->parent->weight);
     printf(" topW lhead %x weight is %x \n",ch,(int)lhead->weight);
*/


       while (loc[ch]->parent->weight < 4*(MAX+1)){
       while (lhead->prev)lhead=lhead->prev;
       if(lhead == loc[ch])break;
       increment(lhead,1);}
       while (lhead->prev)lhead=lhead->prev; /* get head */

/*
     printf(" topX baby %x weight is %x \n",ch,(int)loc[ch]->weight);
     printf(" topX parent %x weight is %x \n",ch,(int)loc[ch]->parent->weight);
     printf(" topX lhead %x weight is %x \n",ch,(int)lhead->weight);
*/

   /* parant weight is 2*(MAX+1) */
   /*  printf(" synbol %x  symbols left = %d \n",loc[ch]->symbol,--YYY); */
     --YYY;
/*
     printf(" did new for %x weight is %x \n",ch,(int)loc[ch]->weight);
     printf(" did parant  weight is %x \n",(int)loc[ch]->parent->weight);
*/
       if((loc[ch]->weight < 2*(MAX+1) || loc[ch]->parent->weight != 4*(MAX+1)) && YYY) 
        {
         printf(" magor error \n");
         exit(1);
         }
       XXX[ch]=1;
       increment(loc[ch],2*(MAX+1));
      
      }
       else
      {

       increment(loc[ch],2*(MAX+1));
       increment(loc[ch],2*(MAX+1));
       }
}
