
/* Adaptive Huffman encoding program */
/* Hal Burch */
/* Sept 16, 1996 */

/* This is based on the Adaptive Huffman algorithm described in Sayood's Data
 * Compression book.  The ranks are not actually stored, but implicitly defined
 * by the location of a node within a doubly-linked list */


/*******************************************************************************
*NOTICE:                                                                       *
*This code is believed by the author to be bug free.  You are free to use and  *
*modify this code with the understanding that any use, either direct or        *
*derivative, must contain acknowledgement of its author and source.  The author*
*makes no warranty of any kind, expressed or implied, of merchantability or    *
*fitness for a particular purpose.  The author shall not be held liable for any*
*incidental or consequential damages in connection with or arising out of the  *
*furnishing, performance, or use of this software.  This software is not       *
*authorized for use in life support devices or systems.                        *
********************************************************************************/
/* modifed by David A. Scott so could be better used for cryptographic uses */
/* #define DEBUG 1 */
#define MAX 255 /* Maximum symbol */
#define N 8 /* number of bits */

#include <stdio.h>
#include <assert.h>
#include <malloc.h>
#include <string.h>
#include "ad2hc.h"

char buff[2*256+1];
int bloc=0;
int outcch=0;
Node *tree;
Node *lhead;
Node *loc[MAX+1];
Node TREE[520];
Node *CP[MAX +1];
Node **CPP;

void flushbuff(FILE *fout)
 { /* Flush buffer */
  fwrite(buff, sizeof(char), (bloc + 7) / 8, fout);
  bloc = 0;
 }

static int FONE;
static int NZZS=0;
static int PC=0;
static int BC=0;
static int PCOLD=0;
static int FZ=0;
static int jj=0;
static int kk=0;


void add_bit (char bit, FILE *fout)
 { /* Add a bit to the output file (buffered) */
  int nc,nx;
  assert(bit == 0 || bit == 1);
  assert( bloc/8 <= 2*128);
  nx = (bloc - NZZS)/8;
  if (bloc%8 == 0 && nx > 0)
   {
    fwrite (buff, sizeof(char), nx , fout);
    memmove (buff, buff+nx, sizeof(buff)-nx);
    bloc -= nx*8;
   }
  buff[bloc/8 +1]=0;
  buff[bloc/8] |= bit << (7-(bloc % 8));
#ifdef WATCH
  if (bit == 1) printf (".");
  else printf (" ");
#endif
/** code from h2unc.c **/

 
  jj = bloc%8; /* the current bit from 0 to 7 on byte */
  if ( FONE ) 
    {  /* get final status of previous huffman code */
      kk = NZZS ==  jj ;  /* true if trailing ones on prevoius code */
      if ( (BC > 8) && kk && (jj != 0))
         PCOLD = PC || PCOLD ||  FZ == 8 || BC > 16;
      else
         PCOLD = 0; /* zero if not a cuttoff one condition */
     FZ = bit == 0 ?1:0;
     PC = bit;
     BC = jj+1; /* bit counter */
    } 
  else { 
    BC++;
    if (BC < 9 ){
      PC |= bit;  /* if a one in first byte */
      if ( bit==0 ) FZ++;
      }
  }
 
/** end of code **/




 if ( FONE == 1){ NZZS = 0; FONE =0;}
 if ( bloc%8 == 0) NZZS=0;
 bloc++;
 }
void send(Node *node, Node *child, FILE *fout)
 { /* Send the prefix code for this node */
  if (node->parent)
    send(node->parent, node, fout);
  if (child)
   {
    if (node->right == child)
     {
      add_bit(1, fout);
      NZZS++;
     }
    else 
     {
      assert(node->left == child);
      add_bit(0, fout);
      NZZS=0;
     }
   }
 }

void transmit (int ch, FILE *fout)
 { /* Send a symbol */
    FONE = 1; /* ture when first character */
    send(loc[ch], NULL, fout);
   /* NZZS ready for next character if there is one */
   /* LASTONE = 1 there is zeros that could be padding */

/*  if (NMAX == 9){
  printf(" here it is n ones %x %x\n",ch,loc[ch]->weight);
 
    checkranks(lhead->next); 
printf("\n **dumptree*** \n");
    tree= &TREE[0];
    dumptree(tree);
printf("\n **dumplist*** \n");
    dumplist(lhead);
printf("\n **dumpend*** \n");
   exit(0);
} */


 }
int i,j,k;

void main(int argc, char **argv)
 {
  FILE *fin;
  FILE *fout;
  int ch;
  int cch;
  int n;
  if (argc != 3)
   {
    usage(argv);
    exit(3);
   }

  if ((fin = fopen(argv[1], "rb")) == NULL)
   {
    fprintf (stderr, "Cannot open input file.\n");
    usage(argv);
    exit(3);
   }

  if ((fout = fopen(argv[2], "wb")) == NULL)
   {
    fprintf (stderr, "Cannot open output file.\n");
    usage(argv);
    exit(3);
   }
n= 256-1; /* number of leaf nodes -1 */
CPP=&CP[0];
 for(i = n,j=0; i<=2*n; i++,j++){
 TREE[i].parent= &TREE[(i-1)/2];
  if( i & 0x1)
  TREE[(i-1)/2].right= &TREE[i];
 else
  TREE[(i-1)/2].left= &TREE[i];
 TREE[i].next= &TREE[i-1];
 TREE[i].prev= &TREE[i+1];
 TREE[i].weight=   1;
 TREE[i].right=NULL;
 TREE[i].left=NULL;
 TREE[i].head=CPP;
 loc[j]=&TREE[i];
 TREE[i].symbol= j;
}  
 *CPP++ = &TREE[n]; 
 lhead=&TREE[2*n]; 
 TREE[2*n].prev=NULL;
  /*  TREE[n].next=NULL;
 /*   checkranks(lhead->next);exit (0); */
for(j=n-1; j!=0; j=(j/2)-1){
for(i=j,k=j/2; i>=k; i--){
 TREE[i].parent= &TREE[(i-1)/2];
  if( i & 0x1)
  TREE[(i-1)/2].right= &TREE[i];
 else
  TREE[(i-1)/2].left= &TREE[i];
 TREE[i].next= &TREE[i-1];
 TREE[i].prev= &TREE[i+1];
 TREE[i].weight= TREE[i].right->weight +TREE[i].left->weight;
 TREE[i].head=CPP;
 TREE[i].symbol= INTERNAL_NODE;
}  
  *CPP++ = &TREE[k]; 
}

 TREE[0].parent= NULL;
 TREE[0].next= NULL;
 TREE[0].prev= &TREE[1];
 TREE[0].weight= TREE[0].right->weight +TREE[0].left->weight;
 TREE[0].head=CPP;
 TREE[0].symbol= INTERNAL_NODE;
    tree= &TREE[0];
 *CPP++ = &TREE[0];
/*
    checkranks(lhead->next); 
printf("\n **dumptree*** \n");
    dumptree(tree);
printf("\n **dumplist*** \n");
    dumplist(lhead);
printf("\n **dumpend*** \n");
   exit(0);
*/

  cch = 0;
  while ((ch = getc(fin)) != EOF)
   {
    cch++;
 /*   if (cch % 1024 == 0)
     {
      printf (".");
      fflush (stdout);
     }
*/

/* Transmit symbol */
    transmit(ch, fout);
    
/* Do update */
    add_ref(ch);

#ifdef DEBUG
    checkranks(lhead->next);
#endif
#ifdef TRACE    
    dumptree(tree);
    dumplist(lhead);
#endif
   }
/** code from h2unc.c **/

 
  jj = bloc%8; /* the current bit from 0 to 7 on byte */
      /* get final status of previous huffman code */
      kk = NZZS ==  jj ;  /* true if trailing ones on prevoius code */
      if ( (BC > 8) && kk && (jj != 0))
         PCOLD = PC || PCOLD ||  FZ == 8 || BC > 16;
      else
         PCOLD = 0; /* zero if not a cuttoff one condition */
/** if PCOLD = 0 then not cut off condition  **/
/** end of code **/
if ( PCOLD )
{
 bloc = (bloc >>3) << 3;
}
else
{
  if( bloc%8 !=0 ) add_bit(0,fout);
  if( bloc%8 !=0 ) add_bit(0,fout);
  if( bloc%8 !=0 ) add_bit(0,fout);
  if( bloc%8 !=0 ) add_bit(0,fout);
  if( bloc%8 !=0 ) add_bit(0,fout);
  if( bloc%8 !=0 ) add_bit(0,fout);
  if( bloc%8 !=0 ) add_bit(0,fout);
}
  
  assert( bloc%8 == 0 ); 
  flushbuff(fout);
  printf(" program h2com.exe version 20000110\n");
  printf ("\n");
  fclose(fin);
  fclose(fout);  
  exit(7);
/*
  printf(" here it is n ones %x %x\n",ch,loc[ch]->weight);

    checkranks(lhead->next); 
printf("\n **dumptree*** \n");
    tree= &TREE[0];
    dumptree(tree);  /**/
 }
